module emop.emop [Impl] { 
   main :-
      Impl:main.
},

module Impl [Impl, Action, Output, Modules] {
   constructor: error/2.

   set_action(NewAction) :-
      (
         data.var:check(Action)
      ->
         Action = NewAction
      ;
         throw(
            error(
               'There cannot be more than one action',
               'cmdline'
            )
         )
      ).

   set_output(New) :-
      (
         'data.var':check(Output)
      ->
         Output = New
      ;
         throw(
            error(
               'There cannot be more than one output filename',
               'cmdline'
            )
         )
      ).

   main :-
      catch(
         (
            Impl:parse_arguments
         ),
         E,
         (
            'emop.message':error(E),
            'sys.process':abort
         )
      ),
      catch(
         (
            (
               Action = 'build'
            ->
               emop.compile:build(Modules, Output)
            ;
               Action = 'compile'
            ->
               emop.compile:compile(Modules, Output)
            ;
%               Action = 'test'
%            ->
%               emop.compile:test(Modules, Output)
%            ;
               Action = 'top'
            ->
               emop.top:main
            ;
               Action = 'version'
            ->
               Impl:version
            ;
               Action = 'help'
            ->
               Impl:help
            ;
               Action = 'where'
            ->
               io.std:print_endline(emop.compile:runtime:get)
            ),
            'sys.process':halt
         ),
         E,
         (
            'emop.message':error(E),
            'sys.process':abort
         )
      ).

   version :-
      'io.std':(
         print_endline('EMoP version 0.02'),
         print_endline('Copyright (C) 2008-2011 INRIA, Institut National de Recherche en'),
         print_endline('Informatique et en Automatique, France.'),
         endline,
         print_endline('This is free software.  You may redistribute copies of it under'),
         print_endline('the terms of the GNU General Public Licence'),
         print_endline('<http://www.gnu.org/licences/gpl.html>.'),
         endline,
         print_endline('There is NO WARRANTY, to the extent permitted by law.'),
         endline,
         print_endline('Written by Thierry Martinez <Thierry.Martinez@inria.fr>.')
      ).

   help :-
      ProgramName = 'sys.cmdline':({0}),
      'io.std':(
         print_endline('Usage:\n'),
         format(
            '~a [build] <modules>...\n\tBuild given modules.\n\n',
            [ProgramName]),
         format(
            '~a compile <modules>...\n\tCompile (do not link) given modules.\n\n',
            [ProgramName]),
         format(
            '~a test <module>...\n\tTest given modules.\n\n',
            [ProgramName]),
         format(
            '~a version\n\tPrint version informations and exit.\n\n',
            [ProgramName]),
         format(
            '~a help\n\tPrint this help and exit.\n\n',
            [ProgramName]),
         print_endline('Other options:'),
         print_endline('--output-filename=FILENAME, -o FILENAME \tSet output filename.'),
         print_endline('--verbose \tVerbose compilation.'),
         print_endline('--debug, -g \tPrint execution trace.'),
         print_endline('--deps \tPrint dependencies.'),
         print_endline('--runtime=FILENAME \tUse FILENAME as runtime.'),
         print_endline('--local-size=N  \tSet default local  stack size to N kb.'),
         print_endline('--global-size=N \tSet default global stack size to N kb.'),
         print_endline('--trail-size=N  \tSet default trail  stack size to N kb.'),
         print_endline('--cstr-size=N   \tSet default cstr   stack size to N kb.'),
         print_endline('--fixed-sizes \tDo not consult environment variables at run-time.'),
         print_endline('--recompile \tRecompile at least the files on the command-line.'),
         print_endline('--all \tRecompile all files.'),
         print_endline('--ignore-warnings \tBuild even if there are warnings.'),
         print_endline('--temp-dir=PATHNAME \tKeep intermediary files in PATHNAME.'),
         print_endline('--build-dir=PATHNAME \tPut inferface and object files in PATHNAME.'),
         print_endline('--no-wam-patch \tDo not patch WAM files.'),
         print_endline('-P OPTION \t Pass OPTION to the Prolog compiler.'),
         endline
      ).

   parse_arguments :-
      RevModulesRef = 'data.ref.backtrackable':new([]),
      PrologOptionsRef = 'data.ref.backtrackable':new([]),
      catch(
      sys.cmdline:parse() {
         short('h', _) :-
            set_action('help').

         short('v', _) :-
            set_action('version').

         short('c', _) :-
            set_action('compile').

         short('o', Next) :-
            set_output(Next:get).

         short('P', Next) :-
            PrologOptionsRef:set([Next:get | PrologOptionsRef:get]).

         short('g', _) :-
            enable_debug.

         enable_debug :-
            emop.compile:lookup_file('emop/runtime_debug.o', NewRuntime):close,
            emop.compile:runtime:set(NewRuntime).

         action('build').
         action('compile').
%         action('test').
         action('help').
         action('version').
         action('where').

         non_ambiguous(Atom, Set, FullAtom) :-
            String = data.atom:string(Atom),
            Set:get(FullAtom),
            data.list:append(String, _, data.atom:string(FullAtom)),
            !,
            (
               Set:get(OtherAtom),
               OtherAtom \= FullAtom,
               data.list:append(String, _, data.atom:string(OtherAtom))
            ->
               throw(error(data.atom:format('~a is ambiguous', [Atom]), 'cmdline'))
            ;
               true
            ).

         long(Long, Next) :-
            Full = non_ambiguous(Long) {
               get('output-filename').
               get('local-size').
               get('global-size').
               get('trail-size').
               get('cstr-size').
               get('fixed-sizes').
               get('debug').
               get('runtime').
               get('temp-dir').
               get('build-dir').
               get('recompile').
               get('all').
               get('ignore-warnings').
               get('verbose').
               get('deps').
               get('no-wam-patch').
               get(Item) :-
                  action(Item).
            },
            Full match (
               action
            ->
               set_action(Full)
            ;
               'output-filename'
            ->
               set_output(Next:get)
            ;
               'local-size'
            ->
               emop.compile:size('local'):set(data.atom:number(Next:get))
            ;
               'global-size'
            ->
               emop.compile:size('global'):set(data.atom:number(Next:get))
            ;
               'trail-size'
            ->
               emop.compile:size('trail'):set(data.atom:number(Next:get))
            ;
               'cstr-size'
            ->
               emop.compile:size('cstr'):set(data.atom:number(Next:get))
            ;
               'fixed-sizes'
            ->
               PrologOptionsRef:set(['--fixed-sizes' | PrologOptionsRef:get])
            ;
               'runtime'
            ->
               emop.compile:runtime:set(Next:get)
            ;
               'temp-dir'
            ->
               emop.compile:temp_dir(Next:get)
            ;
               'build-dir'
            ->
               emop.compile:set_build_dir('some'(Next:get))
            ;
               'debug'
            ->
               enable_debug
            ;
               'recompile'
            ->
               emop.compile:set_use_cache('no_cmdline')
            ;
               'all'
            ->
               emop.compile:set_use_cache('never')
            ;
               'ignore-warnings'
            ->
               emop.compile:set_do_not_build_with_warning('false')
            ;
               'verbose'
            ->
               emop.compile:set_verbose('true')
            ;
               'deps'
            ->
               emop.compile:set_deps('true')
            ;
               'no-wam-patch'
            ->
               emop.compile:set_wam_patch('false')
            ).

         default(NewAction) :-
            data.var:check(Action),
            Action = non_ambiguous(NewAction) {
               get(Item) :-
                  action(Item).
            }.

         file(Filename) :-
            (
               'data.atom':sub(Filename, _, _, 0, '.mpl')
            ->
               Module = 'file'(Filename),
               RevModulesRef:set([Module | RevModulesRef:get])
            ;
               Ext = ('.pl'; '.o'; '.c'; '.a'; '.wam'; '.fd'),
               'data.atom':sub(Filename, _, _, 0, Ext)
            ->
               PrologOptionsRef:set([Filename | PrologOptionsRef:get])
            ;
               Module = 'name'(Filename, [], '.'),
               RevModulesRef:set([Module | RevModulesRef:get])
            ).
      },
      'error'('cmdline'(Error)),
      (
         Msg = Error match (
            'option_without_argument'(Opt)
         ->
            data.atom:format('Option without argument ~a', [Opt])
         ;
            'invalid_option'(Opt)
         ->
            data.atom:format('Invalid option ~a', [Opt])
         ),
         throw(error(Msg, 'cmdline'))
      )
      ),
      Modules = data.list:reverse(RevModulesRef:get),
      (
         Modules = [],
         Action = 'top'
      ->
         true
      ;
         true
      ),
      emop.compile:prolog_options:set(data.list:reverse(PrologOptionsRef:get)),
      (
         data.var:check(emop.compile:runtime:get)
      ->
         emop.compile:lookup_file('emop/runtime.o', Runtime):close,
         emop.compile:runtime:set(Runtime)
      ;
         true
      ),
      emop.compile:temp_dir(_).
}
