module 'emop.message' [HasWarningRef] { 
   constructor: pos/4, error/2, system_error/1,cmdline/0, file/1, file/2.

   has_warning :-
      HasWarningRef:get('true').

   reset_warning :-
      HasWarningRef:set('false').

   line_text(file(Pb, Pe), Initial, Text) :-
      Pb = pos(_Filename0, Off0, Line0, BOL0),
      Pe = pos(_Filename1, Off1, Line1, BOL1),
      data.number:succ(Line0, SLine0),
      data.number:'-'(Off0, BOL0, C0),
      data.number:'-'(Off1, BOL1, C1),
      (
         Line0 = Line1
      ->
         (
            Initial = 'uppercase'
         ->
            data.atom:format(
               'Line ~d, from character ~d to ~d',
               [SLine0, C0, C1],
               Text
            )
         ;
            Initial = 'lowercase'
         ->
            data.atom:format(
               'line ~d, from character ~d to ~d',
               [SLine0, C0, C1],
               Text
            )
         )
      ;
         data.number:succ(Line1, SLine1),
         (
            Initial = 'uppercase'
         ->
            data.atom:format(
               'From line ~d character ~d to line ~d character ~d',
               [SLine0, C0, SLine1, C1],
               Text
            )
         ;
            Initial = 'lowercase'
         ->
            data.atom:format(
               'from line ~d character ~d to line ~d character ~d',
               [SLine0, C0, SLine1, C1],
               Text
            )
         )
      ).

   position_text(file(Pb, Pe), Text) :-
      catch(
         (
      Pb = pos(Filename0, _Off0, _Line0, _BOL0),
      Pe = pos(Filename1, _Off1, _Line1, _BOL1),
      (
         Filename0 = Filename1
      ->
         true
      ;
         throw('error_across_two_files')
      ),
      (
         Filename0 = ''
      ->
         Text = line_text(file(Pb, Pe), 'uppercase')
      ;
         data.atom:format('File "~a", ~a', [Filename0, line_text(file(Pb, Pe), 'lowercase')], Text)
      )
      ),
      _,
      Text = 'At unknown position'
      ).

   format(cmdline, Message, Kind) :-
      'io.err':format('[~a] ~a\n', [Kind, Message]).

   format(file(File), Message, Kind) :-
      (
         File = ''
      ->
         true
      ;
         io.err:format('File ~a:\n', [File])
      ),
      io.err:format('[~a] ~a\n\n', [Kind, Message]).

   format(file(Pb, Pe), Message, Kind) :-
      'io.err':format(
         '~a:\n[~a] ~a\n\n',
         [
            'emop.message':position_text(file(Pb, Pe)),
            Kind,
            Message
         ]
      ).

   error(Error) :-
      (
         Error = error(Message, Pos),
         format(Pos, Message, 'Error')
      ->
         true
      ;
         Error = error(system_error(Message), _),
         io.err:format('System error: ~a.\n', [Message])
      ->
         true
      ;
         'io.err':format('\n[exception] ~p\n\n', [Error])
      ).

   warning(Message, Pos) :-
      format(Pos, Message, 'Warning'),
      HasWarningRef:set('true').
},
HasWarningRef = data.ref.non_backtrackable:new('false')
